/*
     Copyright (C) 2004, 2020 Hubert Vansteenhuyse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "subnet.h"
char ipbuf[IP_TEXT_WIDTH];

char *ip4_bad[] = {
    "0.0.0.0",
    "127.0.0.1",
    "255.255.255.255",
    0
};

//routine always returns the 1st object,
//the caller must call local_ip_last to retrieve the appended object
//li should be 0 to start the 1st object
local_ip* local_ip_append(local_ip *li) {
    local_ip *apli = local_ip_last(li);
    if(!apli) {
        apli = malloc(sizeof(local_ip));
        if(apli) memset(apli, 0, sizeof(local_ip));
        li = apli;
    } else {
        apli->next = malloc(sizeof(local_ip));
        if(apli->next) memset(apli->next, 0, sizeof(local_ip));
    }
    return li;
}

local_ip* local_ip_free(local_ip *li) {
    local_ip *lilast = li, *linext;
    while(lilast) {
        linext = lilast->next;	//save next before freeing last
        free(lilast);
        lilast = linext;
    }

    return 0;
}

local_ip* local_ip_last(local_ip *li) {
    local_ip *last = li;
    while(last) {
        if(!last->next) return last;
        last = last->next;
    }
    return last;
}

/*
 * tricky thing to find ip addresses on linux and on windows
 * therefore it is disabled by default unless --ip is specified in the command options.
 * the algorithms are strikingly identical on both systems
*/
int get_local_ip(local_ip *out) {
    int				badip, sw;

    if(!out) return 0;
    out->swValid = 0;

#ifdef __LINUX__
    //http://man7.org/linux/man-pages/man3/getifaddrs.3.html
    struct ifaddrs *ifaddr, *ifnext;

    if (getifaddrs(&ifaddr) != -1) {
        ifnext = ifaddr;
        while(!out->swValid && ifnext) {
            if(ifnext->ifa_addr->sa_family == AF_INET) {
                out->address[0] = 0;
                getnameinfo(ifnext->ifa_addr, sizeof(struct sockaddr_in), out->address, 4 * 4, NULL, 0, NI_NUMERICHOST);
                badip = 0;
                badip = 0;
                sw = 0;
                if(out->address[0]) {
                    while(!sw && ip4_bad[badip]) {
                        if( strncmp(out->address, ip4_bad[badip], strlen(ip4_bad[badip]) ) == 0)
                            sw = 1;
                        badip++;
                    }

                    if(!sw) {
                        out->swValid = 1;
                        getnameinfo(ifnext->ifa_netmask, sizeof(struct sockaddr_in), out->mask, 4 * 4, NULL, 0, NI_NUMERICHOST);
                    }

                }

            }
            ifnext = ifnext->ifa_next;
        }
        freeifaddrs(ifaddr);
    }
#endif

#ifdef __WINDOWS__
    ULONG ailen = sizeof(IP_ADAPTER_INFO);
    IP_ADAPTER_INFO *ai = malloc(ailen), *ainext;
    IP_ADDR_STRING	*addressnext;
    if(ai && GetAdaptersInfo(ai, &ailen) == ERROR_BUFFER_OVERFLOW)
        ai = realloc(ai, ailen);

    if(ai && GetAdaptersInfo(ai, &ailen) == NO_ERROR) {
        ainext = ai;
        while(!out->swValid && ainext) {
            addressnext = &ainext->IpAddressList;
            while(!out->swValid && addressnext) {
                badip = 0;
                sw = 0;
                if(	addressnext->IpAddress.String[0] ) {
                    while(!sw && ip4_bad[badip]) {
                        if( strncmp(addressnext->IpAddress.String, ip4_bad[badip], strlen(ip4_bad[badip]) ) == 0)
                            sw = 1;
                        badip++;
                    }

                    if(!sw) {
                        sprintf(out->address, "%s", addressnext->IpAddress.String);
                        sprintf(out->mask, "%s", addressnext->IpMask.String);
                        out->swValid = 1;
                    }

                }

                addressnext = addressnext->Next;
            }
            ainext = ainext->Next;
        }
    }
    free(ai);
#endif

    return out->swValid;
}

long mask_inverse(long msk) {
    return 0xFFFFFFFF - msk;
}

int isvalidmask_inverse(long imask) {
    return isvalidmask(mask_inverse(imask));
}

//returns class ABCDE of mask
char getipclass(long ip){
    if ((ip & 0x80000000) == 0)
        return 'A';		//first bits class A = 0
    if ((ip & 0xC0000000) == 0x80000000)
        return 'B';		//first bits class B = 10
    if ((ip & 0xE0000000) == 0xC0000000)
        return 'C';		//first bits class C = 110
    if ((ip & 0xF0000000) == 0xE0000000)
        return 'D';		//first bits class D = 1110
    return 'E';
}


//returns the amount of bits used in a mask
int bits_mask(long mask){
    int bits = 31;

    while( bits > -1 && (mask & ((long)(1 << bits))) )
        bits--;

    return 31 - bits;
}

//returns the amount of subnet bits used
int bits_subnet(void){
    int mbits;
    mbits = bits_mask(mask);
    switch ((int) getipclass(ip)) {
    case (int) 'A':
        return mbits - 8;
        break;
    case (int) 'B':
        return mbits - 16;
        break;
    case (int) 'C':
        return mbits - 24;
        break;
    }
    return -1;
}


//returns the amount of subnets
long getnets(void){
    int sb = bits_subnet();
    return (sb == 0 ? 1 : (long) pow(2, sb) - 2);
}
//puts specified host bits on 1
long hosts_on(int bits){
    long retval = 0;
    while (bits) {
        retval += 1 << (long)(bits - 1);
        bits--;
    }
    return retval;
}

//returns the amount of hosts
long gethosts(void){
    int net = (1 + (int) getipclass(ip) - (int) 'A') * 8;
    if (net > 24)
        return -1;
    return (long) pow(2, 32 - net - bits_subnet()) - 2;
}

//calculates the broadcast address of the specified network
long ipcalcbcast(long network){
    int mbits = bits_mask(mask);
    int subnetbits = bits_subnet();
    int networkbits = mbits - subnetbits;
    int hostbits = 32 - (networkbits + subnetbits);
    return ipcalcnet(network) | hosts_on(hostbits);
}

//calculates the network address of the specified network
long ipcalcnet(long network){
    int mbits = bits_mask(mask);	//#bits mask
    int subnetbits = bits_subnet();	//#bits subnet
    int networkbits = mbits - subnetbits;	//#bits net, ip class (8, 16 of 24)
    int hostbits = 32 - (networkbits + subnetbits);	//#bits hosts
    long basenet = ip & cidrtomask(networkbits);	//calculate network
    long subnet = (network - (subnetbits == 0 ? 1 : 0)) << hostbits;	//build net mask for requested network
    return basenet | subnet;	//add the network bits
}

//returns the ip notation (eg 127.0.0.1) from a long
char *longtoip(long ip){
    int w, x, y, z;
    w = (ip >> 24) & 0xFF;
    x = (ip >> 16) & 0xFF;
    y = (ip >> 8) & 0xFF;
    z = ip & 0xFF;
    sprintf(ipbuf, "%d.%d.%d.%d", w, x, y, z);
    return ipbuf;
}

//returns the inverse ip notation (eg 127.0.0.1) from a long
char *longtoip_inverse(long ip){
    long imsk = mask_inverse(ip);
    sprintf(ipbuf, "%d.%d.%d.%d", getipdigit(imsk, 4), getipdigit(imsk, 3), getipdigit(imsk, 2),getipdigit(imsk, 1));
    return ipbuf;
}

//returns the long for an ip eg: 127.0.0.1
long iptolong(char *ip){
    char tmp[16];
    long ipa = 0;
    int len = (int) strlen(ip), byte, bcnt = 4 * 8, to = 0;
    while (to < len) {
        memset(tmp, 0, 16);
        while (to < len) {
            if (ip[to] == '.') break;
            tmp[strlen(tmp)] = ip[to];
            to++;
        }
        bcnt -= 8;
        byte = atoi(tmp);
        ipa += (long) (byte << bcnt);
        to++;
    }
    return ipa;
}


//returns the netmask for a cidr mask (eg: 8 = 255.0.0.0 converted to its long value
long cidrtomask(int thecidrmask){
    int _thecidrmask = thecidrmask;
    long retval = 0;

    if (thecidrmask < 8 || thecidrmask > 30)
        return retval;

    while (thecidrmask) {
        retval += (long) pow(2, thecidrmask);
        thecidrmask--;
    }
    return retval << (31 - _thecidrmask);
}


//returns a string like nnnnnnn.nnnnnnn.ssshhhhhhh,hhhhh... based on the mask
char binrepresentation[64];
char *getipnetbits(void){
    int nbits = bits_mask(mask), cnt = 0, c, from = 0;
    memset(binrepresentation, 0, 64);
    for (c = 0; c < 32; c++) {
        sprintf(binrepresentation + strlen(binrepresentation), "%c",
                (nbits > 0 ? 'n' : 'h'));
        cnt++;
        if (cnt > 7 && c < 31) {
            sprintf(binrepresentation + strlen(binrepresentation), ".");
            cnt = 0;
        }
        nbits--;
    }
    nbits = bits_subnet();
    cnt = 0;
    switch ((int) getipclass(ip)) {
    case (int) 'A':
        from = 8;
        cnt = 1;
        break;
    case (int) 'B':
        from = 16;
        cnt = 2;
        break;
    case (int) 'C':
        from = 24;
        cnt = 3;
        break;
    }
    if (from > 0) {
        for (c = 0; c < nbits; c++) {
            if (binrepresentation[from + c + cnt] == '.')
                cnt++;
            binrepresentation[from + cnt + c] = 's';
        }
    }
    return binrepresentation;
}

//returns 10101010.10101010.10101010.10101010 for an ip / netmask / reverse / netmask
char* getiptobin(char*ip) {
    static char tmp[16], tmpip[64];
    long lip;
    int n,d, p;
    memset(binrepresentation, 0, 64);
    if(ip && strlen(ip) < 16) {
        memset(tmp, 0, 16);
        memset(tmpip, 0, 64);
        sprintf(tmpip, "%s", ip);
        lip = iptolong(tmpip);
        for(n = 0; n < 4; n++) {
            if(n > 0 && n < 4) sprintf(binrepresentation + strlen(binrepresentation), ".");

            d = getipdigit(lip, n + 1);
            if(d < 0 || d > 255) {
                sprintf(binrepresentation + strlen(binrepresentation), "?");
            } else {
                memset(tmp, '0', 8);
                p = 7;
                while(d) {
                    if(d % 2)tmp[p] = '1';
                    d /= 2;
                    p--;
                }

                sprintf(binrepresentation + strlen(binrepresentation), "%s", tmp);
            }

        }
    }
    return binrepresentation;
}

//returns the digit of an ip address. eg: digit 2 of 193.19.136.1 = 19
int getipdigit(long ipaddress, int digit){
    int dig = 0;
    if (digit < 1 || digit > 4) return 0;
    digit = 32 - (digit  * 8);
    dig = ipaddress >> digit;
    return dig & 0xFF;
}


//returns the magic number (= 256 minus LAST NONZERO NETMASK BYTE)
int magicnumber(void){
    int netbits = bits_mask(mask),
        ipdigit = (netbits / 8)  + (netbits % 8 ? 1 : 0),
        magic = getipdigit(mask, ipdigit);
    return (magic == 0 ? 1 : 256 - magic);
}


//returns the private/public, apipa or loopback type
char *iptype(int ipaddress){
    if (getipdigit(ipaddress, 1) == 127)
        return "loopb";
    if (getipdigit(ipaddress, 1) == 10)
        return "priv";
    if (getipdigit(ipaddress, 1) == 172 && getipdigit(ipaddress, 2) >= 16
            && getipdigit(ipaddress, 2) <= 31)
        return "priv";
    if (getipdigit(ipaddress, 1) == 192 && getipdigit(ipaddress, 2) == 168)
        return "priv";
    if (getipdigit(ipaddress, 1) == 169 && getipdigit(ipaddress, 2) == 254)
        return "apipa";
    return "pub";
}



/*checks the ip format and value. returns 1 for ok, 0 for not ok */
int isvalidip(char *theip){
    int x, pcnt = 0;
    long newip;
    char ipdig[32];
    memset(ipdig, 0, 32);
    for (x = 0; x <= (int) strlen(theip); x++) {
        if (ISDIGIT(theip[x]))
            ipdig[strlen(ipdig)] = theip[x];

        else {
            if (strlen(ipdig) == 0)
                return 0;
            if (theip[x] != '.' && theip[x] != 0) return 0;
            if (atoi(ipdig) > 255 || atoi(ipdig) < 0) return 0;
            pcnt++;
            if (pcnt > 4) return 0;
            memset(ipdig, 0, 32);
        }
        if (strlen(ipdig) > 3) return 0;
    }
    newip = iptolong(theip);
    if (newip == 0 ||
            newip == -1 ||
            getipdigit(newip, 1) == 0 ||
            getipdigit(newip, 1) == 255 ||
            getipdigit(newip, 4) == 0 ||
            getipdigit(newip, 4) == 255
       ) return 0;
    return 1;
}


/*checks the subnet mask value. returns 1 for ok, 0 for not ok */
int isvalidmask(    long themask){
    int mbits, b = 0;
    if (themask == 0 || themask == -1)
        return 0;
    mbits = bits_mask(themask);
    if (mbits > (32 - 2))
        return 0;

    //test consecutive bits with value 0
    while( !(themask & ((unsigned long)(1 << b))) ) 	b++;

    //test consecutive bits with value 1
    while( (themask & ((unsigned long)(1 << b))) ) 	b++;

    //if all the bits were not tested then invalid mask. ok = 111110000000... error: 1111100001000...
    if (b != 32)
        return 0;

    //test the mask for the current ip class
    switch ((int) getipclass(ip)) {
    case (int) 'A':
        if (mbits == 8)
            return 1;
        if (mbits < (8 + 2))
            return 0;
        break;
    case (int) 'B':
        if (mbits == 16)
            return 1;
        if (mbits < (16 + 2))
            return 0;
        break;
    case (int) 'C':
        if (mbits == 24)
            return 1;
        if (mbits < (24 + 2))
            return 0;
        break;
    }
    return 1;
}

//sets mask for nearest amount of networks
void useamount_net(long nets){
    int 	mbits = bits_mask(mask),
            subnetbits = bits_subnet(),
            networkbits = mbits - subnetbits,
            maxsubnetbits = 30 - networkbits,
            b = 2;	// minimum # subnetbits = 2

    if (nets > 1) {
        //choose the nearest amount of subnet bits for the wanted amount of networks
        while (b < maxsubnetbits) {
            if ((long) pow(2, b) - 2 >= nets) break;
            b++;
        }
    } else {
        b = 0;
    }

    mask = cidrtomask(networkbits + b);
    wild = mask_inverse(mask);
}


//sets mask for nearest amount of hosts
void useamount_host(long hosts){
    int 	mbits = bits_mask(mask),
            subnetbits = bits_subnet(),
            networkbits = mbits - subnetbits,
            maxhostbits = 30 - networkbits,
            b = 2;	// minimum # hostbits = 2

    while (b <= maxhostbits) {
        if ((long) pow(2, b) - 2 >= hosts) break;
        b++;
    }

    if (b > maxhostbits)
        mask = cidrtomask(networkbits);
    else
        mask = cidrtomask(networkbits + (32 - networkbits) - b);

    wild = mask_inverse(mask);
}


//returns the subnet of an ip
long findnet(long theip){
    int mbits = bits_mask(mask),
        subnetbits = bits_subnet(),
        networkbits = mbits - subnetbits,
        hostbits =  32 - (networkbits + subnetbits);

    long 	basemask = cidrtomask(networkbits),
            basenet = ip & basemask,
            theSubNet = 0;

    if ((theip & basemask) != basenet) return -1;

    theSubNet = theip - basenet;
    theSubNet = theSubNet >> hostbits;

    return theSubNet;
}

